@@ -5,11 +5,11 @@ X9K3
 """
 
 import argparse
+import datetime
 import io
 import os
 import sys
 import time
-from base64 import b64encode
 from collections import deque
 from functools import partial
 from operator import itemgetter
@@ -19,7 +19,7 @@ from iframes import IFramer
 
 MAJOR = "0"
 MINOR = "1"
-MAINTAINENCE = "41"
+MAINTAINENCE = "49"
 
 
 def version():
@@ -29,6 +29,10 @@ def version():
     Odd number versions are releases.
     Even number versions are testing builds between releases.
 
+    Used to set version in setup.py
+    and as an easy way to check which
+    version you have installed.
+
     """
     return f"{MAJOR}.{MINOR}.{MAINTAINENCE}"
 
@@ -59,53 +63,158 @@ class SCTE35:
     def __init__(self):
         self.cue = None
         self.cue_out = None
-        self.cue_tag = None
         self.cue_time = None
+        self.tag_method = self.x_scte35
+        self.break_timer = None
+        self.break_duration = None
+        self.break_auto_return = False
+        self.event_id = 1
 
-    @staticmethod
-    def mk_cue_tag(cue):
+    def mk_cue_tag(self):
         """
         mk_cue_tag
         """
-        if cue:
-            return f'#EXT-X-SCTE35:CUE="{b64encode(cue.bites).decode()}"'
-        return "# No Cue"
+        if self.cue:
+            return self.tag_method()
+        return False
 
-    @staticmethod
-    def is_cue_out(cue):
+    def x_cue(self):
+        """
+        #EXT-X-CUE-( OUT | IN | CONT )
+        """
+        if self.cue_out == "OUT":
+            self.break_timer = 0
+            return f"#EXT-X-CUE-OUT:{self.break_duration}"
+        if self.cue_out == "IN":
+            return "#EXT-X-CUE-IN"
+        if self.cue_out == "CONT":
+            return f"#EXT-X-CUE-OUT-CONT:{self.break_timer:.3f}/{self.break_duration}"
+        return False
+
+    def x_splicepoint(self):
+        """
+        #EXT-X-SPLICEPOINT-SCTE35
+        """
+        base = f"#EXT-X-SPLICEPOINT-SCTE35:{self.cue.encode()}"
+        if self.cue_out == "OUT":
+            return f"{base}"
+        if self.cue_out == "IN":
+            return f"{base}"
+        return False
+
+    def x_scte35(self):
+        """
+        #EXT-X-SCTE35
+        """
+        base = f'#EXT-X-SCTE35:CUE="{self.cue.encode()}" '
+        if self.cue_out == "OUT":
+            return f"{base},CUE-OUT=YES "
+        if self.cue_out == "IN":
+            return f"{base},CUE-IN=YES "
+        if self.cue_out == "CONT":
+            return f"{base},CUE-OUT=CONT"
+        return False
+
+    def x_daterange(self):
+        """
+        #EXT-X-DATERANGE
+        """
+        fbase = f'#EXT-X-DATERANGE:ID="{self.event_id}"'
+        iso8601 = f"{datetime.datetime.utcnow().isoformat()}Z"
+        fdur = ""
+        if self.break_duration:
+            fdur = f",PLANNED-DURATION={self.break_duration}"
+
+        if self.cue_out == "OUT":
+            self.break_timer = 0
+            fstart = f',START-DATE="{iso8601}"'
+            tag = f"{fbase}{fstart}{fdur},SCTE35-OUT={self.cue.encode_as_hex()}"
+            self.event_id += 1
+            return tag
+
+        if self.cue_out == "IN":
+            fstop = f',END-DATE="{iso8601}"'
+            tag = f"{fbase}{fstop},SCTE35-IN={self.cue.encode_as_hex()}"
+            self.event_id += 1
+            return tag
+
+        return False
+
+    def is_cue_out(self, cue):
         """
         is_cue_out checks a Cue instance
         to see if it is a cue_out event.
         Returns True for a cue_out event.
         """
-        if cue.command.command_type == 5:
-            if cue.command.out_of_network_indicator:
-                return True
+        cmd = cue.command
+        if cmd.command_type == 5:
+            if cmd.out_of_network_indicator:
+                if cmd.break_duration:
+                    self.break_duration = cmd.break_duration
+                    self.break_timer = 0
+                    return True
+
+        upid_starts = [
+            0x10,
+            0x20,
+            0x22,
+            0x30,
+            0x32,
+            0x34,
+            0x36,
+            0x38,
+            0x3A,
+            0x3C,
+            0x3E,
+            0x44,
+            0x46,
+        ]
+        if cmd.command_type == 6:
+            for dsptr in cue.descriptors:
+                if dsptr.tag == 2:
+                    if dsptr.segmentation_type_id in upid_starts:
+                        if dsptr.segmentation_duration:
+                            self.break_duration = dsptr.segmentation_duration
+                            self.break_timer = 0
+                            return True
+
         return False
 
-    @staticmethod
-    def is_cue_in(cue):
+    def is_cue_in(self, cue):
         """
         is_cue_in checks a Cue instance
         to see if it is a cue_in event.
         Returns True for a cue_in event.
         """
-        if cue.command.command_type == 5:
-            if not cue.command.out_of_network_indicator:
-                return True
-        return False
+        cmd = cue.command
+        if cmd.command_type == 5:
+            if not cmd.out_of_network_indicator:
+                if self.break_timer >= self.break_duration:
+                    return True
 
-    def cue_out_cue_in(self):
-        """
-        cue_out_cue_in adds CUE-OUT
-        and CUE-IN attributes to hls scte35 tags
-        """
-        if self.is_cue_out(self.cue):
-            self.cue_tag += ",CUE-OUT=YES"
-            self.cue_out = "CONT"
-        if self.is_cue_in(self.cue):
-            self.cue_tag += ",CUE-IN=YES"
-            self.cue_out = None
+        upid_stops = [
+            0x11,
+            0x21,
+            0x21,
+            0x23,
+            0x33,
+            0x35,
+            0x37,
+            0x39,
+            0x3B,
+            0x3D,
+            0x3F,
+            0x45,
+            0x47,
+        ]
+        if cmd.command_type == 6:
+            for dsptr in cue.descriptors:
+                if dsptr.tag == 2:
+                    if dsptr.segmentation_type_id in upid_stops:
+                        if self.break_timer >= self.break_duration:
+                            return True
+
+        return False
 
 
 class X9K3(Stream):
@@ -114,7 +223,6 @@ class X9K3(Stream):
     """
 
     # sliding window size
-    WINDOW_SLOTS = 5
     __slots__ = [
         "_tsdata",
         "start",
@@ -135,23 +243,27 @@ class X9K3(Stream):
         set show_null=False to exclude Splice Nulls
         """
         super().__init__(tsdata, show_null)
+        self._tsdata = tsdata
+        self.in_stream = tsdata
         self.active_segment = io.BytesIO()
         self.active_data = io.StringIO()
         self.start = False
         self.scte35 = SCTE35()
+        self.seg = SegData()
+        self.iframer = IFramer(shush=True)
         self.window = []
+        self.window_size = 5
         self.window_slot = 0
         self.header = None
         self.live = False
         self.output_dir = "."
         self.delete = False
-        self.seg = SegData()
         self.sidecar_file = None
         self.sidecar = None
         self.seconds = 2
         self.replay = False
+        self.discontinuity_sequence = 0
         self._parse_args()
-        self.iframer = IFramer()
 
     def _parse_args(self):
         """
@@ -192,12 +304,18 @@ class X9K3(Stream):
         )
 
         parser.add_argument(
-            "-l",
-            "--live",
-            action="store_const",
-            default=False,
-            const=True,
-            help="Flag for a live event ( enables sliding window m3u8 )",
+            "-T",
+            "--hls_tag",
+            default="x_cue",
+            help="x_scte35, x_cue, x_daterange, or x_splicepoint  (default x_cue)",
+        )
+
+        parser.add_argument(
+            "-w",
+            "--window_size",
+            default=5,
+            type=int,
+            help="sliding window size(default:5)",
         )
 
         parser.add_argument(
@@ -209,6 +327,15 @@ class X9K3(Stream):
             help="delete segments ( enables --live )",
         )
 
+        parser.add_argument(
+            "-l",
+            "--live",
+            action="store_const",
+            default=False,
+            const=True,
+            help="Flag for a live event ( enables sliding window m3u8 )",
+        )
+
         parser.add_argument(
             "-r",
             "--replay",
@@ -242,6 +369,17 @@ class X9K3(Stream):
         else:
             self._tsdata = sys.stdin.buffer
 
+    def _args_hls_tag(self, args):
+        tag_map = {
+            "x_scte35": self.scte35.x_scte35,
+            "x_cue": self.scte35.x_cue,
+            "x_daterange": self.scte35.x_daterange,
+            "x_splicepoint": self.scte35.x_splicepoint,
+        }
+        if args.hls_tag not in tag_map:
+            raise ValueError(f"hls tag  must be in {tag_map.keys()}")
+        self.scte35.tag_method = tag_map[args.hls_tag]
+
     def _args_output_dir(self, args):
         self.output_dir = args.output_dir
         if not os.path.isdir(args.output_dir):
@@ -263,6 +401,9 @@ class X9K3(Stream):
     def _args_time(self, args):
         self.seconds = args.time
 
+    def _args_window_size(self, args):
+        self.window_size = args.window_size
+
     def _apply_args(self, args):
         """
         _apply_args  uses command line args
@@ -270,9 +411,11 @@ class X9K3(Stream):
         """
         self._args_version(args)
         self._args_input(args)
+        self._args_hls_tag(args)
         self._args_output_dir(args)
         self._args_sidecar(args)
         self._args_time(args)
+        self._args_window_size(args)
         self._args_flags(args)
         if isinstance(self._tsdata, str):
             self._tsdata = reader(self._tsdata)
@@ -296,23 +439,26 @@ class X9K3(Stream):
         """
         m3u = "#EXTM3U"
         m3u_version = "#EXT-X-VERSION:3"
-        m3u_cache = "#EXT-X-ALLOW-CACHE:YES"
-        headers = [m3u, m3u_version, m3u_cache]
-        if not self.live:
-            play_type = "#EXT-X-PLAYLIST-TYPE:VOD"
-            headers.append(play_type)
-        target = f"#EXT-X-TARGETDURATION:{int(self.seconds+1)}"
-        headers.append(target)
+        target = f"#EXT-X-TARGETDURATION:{int(self.seconds*2)}"
         seq = f"#EXT-X-MEDIA-SEQUENCE:{self.seg.start_seg_num}"
-        headers.append(seq)
+        dseq = f"#EXT-X-DISCONTINUITY-SEQUENCE:{self.discontinuity_sequence}"
         bumper = ""
-        headers.append(bumper)
-        self.header = "\n".join(headers)
+        self.header = "\n".join(
+            [
+                m3u,
+                m3u_version,
+                target,
+                seq,
+                dseq,
+                bumper,
+            ]
+        )
 
     def load_sidecar(self, file, pid):
         """
         load_sidecar reads (pts, cue) pairs from
         the sidecar file and loads them into X9K3.sidecar
+        if live, blank out the sidecar file after cues are loaded.
         """
         if self.sidecar_file:
             with reader(file) as sidefile:
@@ -322,10 +468,15 @@ class X9K3(Stream):
                         pts, cue = line.split(",", 1)
                         if float(pts) >= self.pid2pts(pid):
                             if [float(pts), cue] not in self.sidecar:
+                                print("loading",pts,cue)
                                 self.sidecar.append([float(pts), cue])
                                 self.sidecar = deque(
                                     sorted(self.sidecar, key=itemgetter(0))
                                 )
+                sidefile.close()
+            if self.live:
+                with open(self.sidecar_file,'w') as scf:
+                    scf.close()
 
     def chk_sidecar_cues(self, pid):
         """
@@ -337,7 +488,7 @@ class X9K3(Stream):
                 raw = self.sidecar.popleft()[1]
                 self.scte35.cue = Cue(raw)
                 self.scte35.cue.decode()
-                self.scte35.cue_tag = self.scte35.mk_cue_tag(self.scte35.cue)
+                # self.scte35.cue_tag = self.scte35.mk_cue_tag(self.scte35.cue)
                 self._chk_cue(pid)
 
     def chk_stream_cues(self, pkt, pid):
@@ -352,8 +503,8 @@ class X9K3(Stream):
 
     def _add_discontinuity(self):
         self.active_data.write("#EXT-X-DISCONTINUITY\n")
-        self.seg.seg_start = None
-        self.seg.seg_stop = None
+    #    self.seg.seg_start = None
+      #  self.seg.seg_stop = None
 
     def _chk_cue(self, pid):
         """
@@ -368,29 +519,31 @@ class X9K3(Stream):
             print(
                 f"Preroll: {round(self.scte35.cue.command.pts_time- self.pid2pts(pid), 6)} "
             )
-            self.scte35.cue_tag = self.scte35.mk_cue_tag(self.scte35.cue)
-            self.scte35.cue_out = None
+            #self.scte35.cue_out = None
 
         else:
             self.scte35.cue_time = self.pid2pts(pid)
             self.active_data.write("# Splice Immediate\n")
-            self._mk_cue_splice_point()
+           # self._mk_cue_splice_point(pid)
 
-    def _mk_cue_splice_point(self):
+    def _mk_cue_splice_point(self,pid):
         """
         _mk_cue_splice_point inserts a tag
         at the time specified in the cue.
         """
         if self.scte35.cue:
-            self.scte35.cue_tag = self.scte35.mk_cue_tag(self.scte35.cue)
-            self._add_discontinuity()
-            print(f"Splice Point {self.scte35.cue.command.name}@{self.scte35.cue_time}")
-            self.active_data.write(f"# Splice Point @ {self.scte35.cue_time}\n")
-            self.scte35.cue_out_cue_in()
+            if self.scte35.cue_time == self.pid2pts(pid):
+                if self.scte35.is_cue_out(self.scte35.cue):
+                    print(f"Splice Point {self.scte35.cue.command.name}@{self.scte35.cue_time}")
+                    self.active_data.write(f"# Splice Point @ {self.scte35.cue_time}\n")
+                    self.scte35.cue_out = "OUT"
+                    self._add_discontinuity()
+            if self.scte35.break_timer >= self.scte35.break_duration:
+                self.active_data.write(f"# Splice Point @ {self.scte35.cue_time}\n")
+                self._add_discontinuity()
+
             if self.scte35.cue_out is None:
                 self.scte35.cue_time = None
-            self.active_data.write(self.scte35.cue_tag + "\n")
-            self.scte35.cue_tag = None
 
     def cue_out_continue(self):
         """
@@ -399,29 +552,45 @@ class X9K3(Stream):
         the live sliding window of segments
         has a tag with CUE-OUT=CONT
         """
-        if self.window_slot > self.WINDOW_SLOTS:
-            self.window_slot = 0
-        if self.scte35.cue_out == "CONT" and self.window_slot == 0:
-            self.scte35.cue_tag = self.scte35.mk_cue_tag(self.scte35.cue)
-            self.scte35.cue_tag += ",CUE-OUT=CONT"
         self.window_slot += 1
+        self.window_slot = self.window_slot % (self.window_size + 1)
 
     def _mk_segment(self, pid):
         """
         _mk_segment cuts hls segments
         """
         now = self.pid2pts(pid)
+        #if now - self.seg.seg_start >= self.seconds:
+
+
+        if self.scte35.cue_out=='CONT':
+            if self.scte35.is_cue_in(self.scte35.cue):
+                if self.scte35.break_timer >= self.scte35.break_duration:
+                    self.seg.seg_stop = now
+                    self.scte35.cue_out ='IN'
+                    self._mk_cue_splice_point(pid)
+                    self.active_data.write(f'#PTS   {self.pid2pts(pid)}\n')
+                    self.scte35.cue = None
+                    self.scte35.cue_time = None
+                    self._write_segment()
+                    self._write_manifest()
+                    return
+        if now >= self.seg.seg_stop:
+            self.seg.seg_stop = now
         if self.scte35.cue_time:
-            if self.seg.seg_start < self.scte35.cue_time < now:
-                self.seg.seg_stop = self.scte35.cue_time
-                self._mk_cue_splice_point()
-                self.scte35.cue_time = None
+            if self.seg.seg_start < self.scte35.cue_time < self.seg.seg_stop:
+                self.scte35.cue_time = self.seg.seg_stop
+            if self.scte35.cue_time == now:
+                if self.scte35.is_cue_out(self.scte35.cue):
+                    self._mk_cue_splice_point(pid)
+                    self.scte35.cue_time = None
         if self.seg.seg_stop:
-            if now >= self.seg.seg_stop:
+            if self.seg.seg_stop <= now:
                 self.seg.seg_stop = now
                 self._write_segment()
                 self._write_manifest()
 
+
     def _write_segment(self):
         """
         _write_segment creates segment file,
@@ -433,16 +602,18 @@ class X9K3(Stream):
         self.seg.seg_uri = self.mk_uri(self.output_dir, seg_file)
         if self.seg.seg_stop:
             self.seg.seg_time = round(self.seg.seg_stop - self.seg.seg_start, 6)
-            if self.live:
-                self.cue_out_continue()
-            if self.scte35.cue_tag:
-                self.active_data.write(self.scte35.cue_tag + "\n")
-                self.scte35.cue_tag = None
+
+            cue_tag = self.scte35.mk_cue_tag()
+            if cue_tag:
+                self.active_data.write(cue_tag + "\n")
+            if self.scte35.break_duration:
+                self.scte35.break_timer += self.seg.seg_time
+
             with open(self.seg.seg_uri, "wb+") as a_seg:
                 a_seg.write(self.active_segment.getbuffer())
                 a_seg.flush()
             del self.active_segment
-            self.active_data.write(f"#EXTINF:{self.seg.seg_time},\n")
+            self.active_data.write(f"#EXTINF:{self.seg.seg_time:.3f},\n")
             self.active_data.write(seg_file + "\n")
             self.seg.seg_start = self.seg.seg_stop
             self.seg.seg_stop += self.seconds
@@ -450,15 +621,36 @@ class X9K3(Stream):
                 (self.seg.seg_num, self.seg.seg_uri, self.active_data.getvalue())
             )
             self.seg.seg_num += 1
+            if self.scte35.cue_out == "OUT":
+                self.scte35.cue_out = "CONT"
+            self.active_data = io.StringIO()
+            self.active_segment = io.BytesIO()
+            self.stream_diff()
+
+            if self.live:
+                self.cue_out_continue()
+
+    def _auto_return(self):
+        if self.scte35.cue_out == 'CONT':
+           if self.scte35.break_timer >= self.scte35.break_duration:
+                self.scte35.cue_out='IN'
+                self.scte35.ad_timer = None
+                self.active_data.write("#auto")
+                self._add_discontinuity()
+            # self.scte35.cue_out_cue_in()
 
     def _pop_segment(self):
-        if len(self.window) > self.WINDOW_SLOTS:
+        if len(self.window) > self.window_size:
             if self.delete:
                 drop = self.window[0][1]
                 print(f"deleting {drop}")
                 os.unlink(drop)
             self.window = self.window[1:]
 
+    def _discontinuity_seq_plus_one(self):
+        if "DISCONTINUITY" in self.window[0][2]:
+            self.discontinuity_sequence += 1
+
     def _open_m3u8(self):
         m3u8_uri = self.mk_uri(self.output_dir, "index.m3u8")
         return open(m3u8_uri, "w+", encoding="utf-8")
@@ -468,19 +660,23 @@ class X9K3(Stream):
         _write_manifest writes segment meta data from
         self.window to an m3u8 file
         """
-        self.stream_diff()
         if self.live:
+            self._discontinuity_seq_plus_one()
             self._pop_segment()
-            self.seg.start_seg_num = self.window[0][0]
+            if self.window:
+                self.seg.start_seg_num = self.window[0][0]
+            else:
+                return
         with self._open_m3u8() as m3u8:
             self._mk_header()
             m3u8.write(self.header)
             for i in self.window:
                 m3u8.write(i[2])
+ #           if self.scte35.cue_out=='IN':
+   #             self.scte35.cue_out = None
             if not self.live:
                 m3u8.write("#EXT-X-ENDLIST")
-        self.active_data = io.StringIO()
-        self.active_segment = io.BytesIO()
+
 
     def stream_diff(self):
         """
@@ -511,7 +707,7 @@ class X9K3(Stream):
         fstart = f"\tstart: {rev}{self.seg.seg_start- self.seg.seg_time:.6f}{res}"
         fdur = f"\tduration: {rev}{self.seg.seg_time:.6f}{res}"
         fdiff = f"\tstream diff: {rev}{round(self.seg.diff_total,6)}{res}"
-        print(f"{furi}{fstart}{fdur}{fdiff}", end='\r')
+        print(f"{furi}{fstart}{fdur}{fdiff}")
         self.seg.init_time = now
         if self.live:
             if self.seg.diff_total > 0:
@@ -533,6 +729,7 @@ class X9K3(Stream):
             pts |= payload[13] >> 1
             prgm = self.pid2prgm(pid)
             self.maps.prgm_pts[prgm] = pts
+            #self.active_data.write(f'#PTS {self.as_90k(pts)}\n')
             if not self.seg.seg_start:
                 self.seg.seg_start = self.as_90k(pts)
                 self.seg.seg_stop = self.seg.seg_start + self.seconds
@@ -571,7 +768,7 @@ class X9K3(Stream):
             for pkt in iter(partial(self._tsdata.read, self._PACKET_SIZE), b"")
         ]
         self._add_discontinuity()
-        self._tsdata.seek(0)
+        self._tsdata = reader(self.in_stream)
         return True
 
     def run(self):
@@ -590,8 +787,6 @@ def cli():
     """
     cli provides one function call
     for running X9K3  with command line args
-    usage: x9k3 [-h] [-i INPUT] [-o OUTPUT_DIR] [-s SIDECAR] [-l] [-d] [-r] [-v]
-
     Two lines of code gives you a full X9K3 command line tool.
 
      from X9K3 import cli
@@ -600,8 +795,8 @@ def cli():
     """
     stuff = X9K3()
     stuff.run()
-    return stuff
 
 
 if __name__ == "__main__":
-    x9=cli()
+    x9k = X9K3()
+    x9k.run()
